import numpy as np

def have_fenv() -> bool: ...
def random_double(size: int, rng: np.random.RandomState) -> np.float64: ...
def test_add_round(size: int, mode: str, rng: np.random.RandomState): ...

def _dd_exp(xhi: float, xlo: float) -> tuple[float, float]: ...
def _dd_log(xhi: float, xlo: float) -> tuple[float, float]: ...
def _dd_expm1(xhi: float, xlo: float) -> tuple[float, float]: ...
